# coding: utf-8
###########################################################################
# Copyright (C) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "15/21/2021"


import functools
import logging

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.gui import qt

from tomwer.core.process.reconstruction.nabu.castvolume import CastVolumeTask
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.volume.volumefactory import VolumeFactory

from ..processingstack import FIFO, ProcessingThread

_logger = logging.getLogger(__name__)


class CastVolumeProcessStack(FIFO, qt.QObject):
    def __init__(self, process_id=None):
        qt.QObject.__init__(self)
        FIFO.__init__(self, process_id=process_id)
        self._dry_run = False

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    def _process(self, data, configuration, callback=None):
        ProcessManager().notify_dataset_state(
            dataset=data,
            process=self,
            state=DatasetState.ON_GOING,
        )
        _logger.processStarted(
            f"start cast volume {str(data.get_identifier())} with parameters {configuration}"
        )

        if isinstance(data, TomwerVolumeBase):
            if data.data_url is not None:
                path = data.data_url.file_path()
            else:
                path = None
        elif isinstance(data, TomwerScanBase):
            path = data.path
        else:
            raise ValueError(
                f"data is expected to be an instance of {TomwerScanBase} or {TomwerVolumeBase}. {type(data)} provided"
            )

        if (
            path is not None
            and isOnLbsram(path)
            and is_low_on_memory(get_lbsram_path()) is True
        ):
            # if computer is running into low memory on lbsram skip it
            mess = f"low memory, skip volume cast {data}"
            ProcessManager().notify_dataset_state(
                dataset=data, process=self._process_id, state=DatasetState.SKIPPED
            )
            _logger.processSkipped(mess)
            if callback is not None:
                callback()
            self.scan_ready(scan=data)
        else:
            self._data_currently_computed = data
            try:
                self._computationThread.init(data=data, configuration=configuration)
            except ValueError as e:
                # initialization can fail (for example for cast volume is there is no volume or be case this will raise an error)
                # then we want to keep the thread active
                self._data_currently_computed = None
                ProcessManager().notify_dataset_state(
                    dataset=data, process=self._process_id, state=DatasetState.SKIPPED
                )
                _logger.processSkipped(f"thread initialization failed. Error is {e}")
                if callback is not None:
                    callback()
                self.scan_ready(scan=data)
            else:
                # need to manage connect before starting it because
                fct_callback = functools.partial(
                    self._end_threaded_computation, callback
                )
                self._computationThread.finished.connect(fct_callback)
                self._computationThread.start()

    def _end_computation(self, data, future_tomo_obj, callback):
        """
        callback when the computation thread is finished

        :param scan: pass if no call to '_computationThread is made'
        """
        assert isinstance(data, TomwerObject)
        FIFO._end_computation(
            self, data=data, future_tomo_obj=future_tomo_obj, callback=callback
        )

    def _end_threaded_computation(self, callback=None):
        assert self._data_currently_computed is not None
        self._computationThread.finished.disconnect()
        if callback:
            callback()
        FIFO._end_threaded_computation(self)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _ProcessingThread(process_id=process_id)


class _ProcessingThread(ProcessingThread, SuperviseProcess):
    """
    Thread use to execute the processing of the axis position
    """

    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        try:
            ProcessingThread.__init__(self, process_id=process_id)
        except TypeError:
            ProcessingThread.__init__(self)
        self._dry_run = False
        self._scan = None
        self._future_tomo_obj = None
        self._volume = None
        self._configuration = None

    @property
    def future_tomo_obj(self):
        return self._future_tomo_obj

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    def init(self, data, configuration):
        if isinstance(data, TomwerVolumeBase):
            self._scan = None
            self._volume = data
        else:
            self._scan = data
            if len(data.latest_vol_reconstructions) < 1:
                raise ValueError(
                    "No reconstructed volume found. did you run the 'nabu volume reconstruction' process ?"
                )
            volume_identifier = data.latest_vol_reconstructions[0]
            self._volume = VolumeFactory.create_tomo_object_from_identifier(
                volume_identifier
            )
        self._data = data
        self._configuration = configuration

    def run(self):
        self.sigComputationStarted.emit()
        cast_volume = CastVolumeTask(
            process_id=self.process_id,
            varinfo=None,
            inputs={
                "volume": self._volume,
                "configuration": self._configuration,
            },
        )
        try:
            cast_volume.run()
        except Exception as e:
            _logger.error(str(e))
            if self._scan is not None:
                # if scan is provided update status because otherwise only the volume state is updated
                ProcessManager().notify_dataset_state(
                    dataset=self._scan,
                    process=self,
                    state=DatasetState.FAILED,
                )
        else:
            if self._scan is not None:
                # if scan is provided update status because otherwise only the volume state is updated
                ProcessManager().notify_dataset_state(
                    dataset=self._scan,
                    process=self,
                    state=DatasetState.SUCCEED,
                )
                self._scan.cast_volume = self._volume.cast_volume
            if cast_volume.outputs.future_tomo_obj is not None:
                self._future_tomo_obj = cast_volume.outputs.future_tomo_obj



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
material for radio and sinogram normalization
"""


__authors__ = [
    "H. Payno",
]
__license__ = "MIT"
__date__ = "25/06/2021"


import functools
import logging

import numpy
import tomoscan.esrf.scan.utils
from processview.core.dataset import DatasetIdentifier
from processview.core.superviseprocess import SuperviseProcess
from silx.io.url import DataUrl
from tomoscan.esrf.scan.utils import get_data
from tomoscan.normalization import Method as NormMethod

import tomwer.version
from tomwer.core.process.task import Task
from tomwer.core.utils.scanutils import data_identifier_to_scan

from .params import SinoNormalizationParams, _ValueCalculationFct, _ValueSource

_logger = logging.getLogger(__name__)


class SinoNormalizationTask(
    Task,
    SuperviseProcess,
    input_names=("data",),
    output_names=("data",),
    optional_input_names=("serialize_output_data",),
):
    """
    Task to define the normalization to apply to a sinogram before reconstructing it with nabu
    """

    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        SuperviseProcess.__init__(self, process_id=process_id)
        self._dry_run = False
        if "configuration" in inputs:
            self.set_configuration(inputs["configuration"])

    def set_properties(self, properties):
        if isinstance(properties, SinoNormalizationParams):
            self._settings = properties.to_dict()
        else:
            self._settings = properties

    def set_configuration(self, configuration: dict) -> None:
        self.set_properties(configuration)

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            return
        extra_infos = scan.intensity_normalization.get_extra_infos()
        extra_infos.update(
            {
                "tomwer_processing_res_code": None,
            }
        )
        scan.intensity_normalization.set_extra_infos(extra_infos)
        params = SinoNormalizationParams.from_dict(self._settings)
        # define the method used to the scan
        scan.intensity_normalization.method = params.method
        # after this processing the source
        try:
            if params.method in (NormMethod.NONE, NormMethod.CHEBYSHEV):
                final_norm_info = {}
            elif params.source is _ValueSource.MANUAL_ROI:
                value = self._compute_from_manual_roi(scan)
                # need_conversion_to_tomoscan = True
                # insure this could be hashable (for caches)
                if isinstance(value, numpy.ndarray):
                    value = tuple(value)
                final_norm_info = {
                    "value": value,
                    "source": _ValueSource.MANUAL_SCALAR.value,
                }
            elif params.source is _ValueSource.AUTO_ROI:
                value = self._compute_from_automatic_roi(scan)
                # need_conversion_to_tomoscan = True
                # insure this could be hashable (for caches)
                if isinstance(value, numpy.ndarray):
                    value = tuple(value)
                final_norm_info = {
                    "value": value,
                    "source": _ValueSource.MANUAL_SCALAR.value,
                }
            elif params.source is _ValueSource.DATASET:
                final_norm_info = {
                    "dataset_url": params.extra_infos.get("dataset_url", None),
                    "source": _ValueSource.DATASET.value,
                }
            elif params.source is _ValueSource.MANUAL_SCALAR:
                final_norm_info = {
                    "value": params.extra_infos.get("value", None),
                    "source": _ValueSource.MANUAL_SCALAR.value,
                }
            else:
                raise ValueError(f"method {params.method} is not handled")
        except Exception as e:
            _logger.error(e)
            final_norm_info = {"tomwer_processing_res_code": False}
        else:
            final_norm_info.update({"tomwer_processing_res_code": True})
        scan.intensity_normalization.set_extra_infos(final_norm_info)
        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan

    def _compute_from_manual_roi(self, scan):
        params = SinoNormalizationParams.from_dict(self.get_configuration())
        extra_info = params.extra_infos
        start_x = extra_info.get("start_x", None)
        end_x = extra_info.get("end_x", None)
        start_y = extra_info.get("start_y", None)
        end_y = extra_info.get("end_y", None)

        calc_fct = extra_info.get("calc_fct", None)
        if calc_fct is None:
            raise ValueError("calc_fct should be provided")
        else:
            calc_fct = _ValueCalculationFct.from_value(calc_fct)

        try:
            value = self._cache_compute_from_manual_roi(
                dataset_identifier=scan.get_identifier(),
                start_x=start_x,
                start_y=start_y,
                end_x=end_x,
                end_y=end_y,
            )[calc_fct.value]
        except Exception as e:
            _logger.error(e)
            return None
        else:
            return value

    @staticmethod
    @functools.lru_cache(
        maxsize=6
    )  # maxsize=6 to at most keep info at volume and at frame level for 3 scans
    def _cache_compute_from_manual_roi(
        dataset_identifier: DatasetIdentifier,
        start_x,
        end_x,
        start_y,
        end_y,
    ) -> dict:
        """
        compute mean and median on a volume or a slice.
        To improve io performances compute both mean and median for both
        cases one per frame or one for the entire volume

        :param projections:
        :param start_x:
        :param end_x:
        :param start_y:
        :param end_y:
        :param calc_area:
        :return:
        :rtype: dict
        """
        if start_x is None or end_x is None or start_y is None or end_y is None:
            raise ValueError("min_x, max_x, min_y and max_y should all be provided.")
        else:
            start_x = int(start_x)
            start_y = int(start_y)
            end_x = int(end_x)
            end_y = int(end_y)
        # clamp ROI with 0 border
        start_x = max(0, start_x)
        start_y = max(0, start_y)
        end_x = max(0, end_x)
        end_y = max(0, end_y)
        scan = dataset_identifier.recreate_dataset()
        projections = scan.projections

        roi_area = numpy.zeros(
            (len(projections), int(end_y - start_y), int(end_x - start_x))
        )
        # as dataset can be large we try to split the load of
        # the data
        proj_compacted = tomoscan.esrf.scan.utils.get_compacted_dataslices(
            projections,
            max_grp_size=20,
        )
        proj_indexes = sorted(proj_compacted.keys())
        # small hack to avoid loading several time the same url
        url_treated = set()

        def url_has_been_treated(url: DataUrl):
            return (
                url.file_path(),
                url.data_path(),
                url.data_slice().start,
                url.data_slice().stop,
                url.data_slice().step,
                url.scheme(),
            ) in url_treated

        def append_url(url: DataUrl):
            url_treated.add(
                (
                    url.file_path(),
                    url.data_path(),
                    url.data_slice().start,
                    url.data_slice().stop,
                    url.data_slice().step,
                    url.scheme(),
                )
            )

        current_idx = 0
        url_idxs = {v.path(): k for k, v in scan.projections.items()}
        for proj_index in proj_indexes:
            url = proj_compacted[proj_index]
            if url_has_been_treated(url):
                continue

            append_url(url)
            data = get_data(url)
            if data.ndim < 2:
                raise ValueError("data is expected to be at least 2D")
            # clamp ROI with frame size
            start_x = min(data.shape[-1], start_x)
            start_y = min(data.shape[-2], start_y)
            end_x = min(data.shape[-1], end_x)
            end_y = min(data.shape[-2], end_y)

            def retrieve_data_proj_indexes(url_):
                urls = []
                for data_slice in range(
                    url_.data_slice().start,
                    url_.data_slice().stop,
                    url_.data_slice().step or 1,
                ):
                    urls.append(
                        DataUrl(
                            file_path=url_.file_path(),
                            data_path=url_.data_path(),
                            scheme=url_.scheme(),
                            data_slice=data_slice,
                        )
                    )

                # try to retrieve the index from the projections else
                # keep the slice index as the frame index (should be the
                # case in most case
                res = []
                for my_url in urls:
                    my_url_path = my_url.path()
                    if my_url_path in url_idxs:
                        res.append(url_idxs[my_url_path])
                    else:
                        _logger.warning(
                            f"unable to retrieve frame index from url {my_url_path}. Take the slice index as frame index"
                        )
                return res

            data_indexes = retrieve_data_proj_indexes(url)
            # apply flat field correction
            if data.ndim == 2:
                projs = (data,)
            else:
                projs = list(data)
            data = scan.flat_field_correction(projs=projs, proj_indexes=data_indexes)
            if data is None:
                continue
            data = numpy.asarray(data)
            if data.ndim == 2:
                roi_area[current_idx] = data[start_y:end_y, start_x:end_x]
                current_idx += 1
            elif data.ndim == 3:
                length = data.shape[0]
                roi_area[current_idx : current_idx + length] = data[
                    :, start_y:end_y, start_x:end_x
                ]
                current_idx += length
            else:
                raise ValueError(f"Frame where expected and not a {data.ndim}D object")
        return SinoNormalizationTask.compute_stats(roi_area)

    @staticmethod
    def compute_stats(data):
        results = {}
        for calc_fct in _ValueCalculationFct:
            if data.ndim == 3:
                res = getattr(numpy, calc_fct.value)(data, axis=(-2, -1))
            elif data.ndim == 2:
                res = getattr(numpy, calc_fct.value)(data, axis=(-1))
            elif data.ndim in (0, 1):
                res = data
            else:
                raise ValueError(f"dataset dimension not handled ({data.ndim}D)")
            results[calc_fct.value] = res
        return results

    def _compute_from_automatic_roi(self, scan):
        raise NotImplementedError("Not implemented yet")

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "Intensity normalization"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "Normalize intensity."



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "25/05/2018"


import logging

from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.control.datalist import GenericScanList
from tomwer.gui.qfolderdialog import QDataDialog

from .selectorwidgetbase import _SelectorWidget

logger = logging.getLogger(__name__)


class ScanSelectorWidget(_SelectorWidget):
    """Widget used to select a scan on a list"""

    def _buildDataList(self):
        return GenericScanList(parent=self)

    def n_scan(self) -> int:
        return super().n_data()

    def _callbackAddData(self):
        dialog = QDataDialog(self, multiSelection=True)

        if not dialog.exec_():
            dialog.close()
            return

        for folder in dialog.files_selected():
            tomo_objs = self.add(folder)
            self.setMySelection(tomo_objs)
        self.sigUpdated.emit()

    def setActiveData(self, data):
        if isinstance(data, TomwerScanBase):
            data = data.path
        super().setActiveData(data)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "09/08/2018"


import io
import json
import logging
import os
import typing
from glob import glob
from typing import Optional
import functools

import numpy
from silx.io.url import DataUrl
from silx.utils.enum import Enum as _Enum
from silx.io.utils import open as open_hdf5

from tomoscan.identifier import VolumeIdentifier
from tomoscan.normalization import IntensityNormalization
from tomoscan.volumebase import VolumeBase
from tomoscan.identifier import BaseIdentifier

from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.utils.ftseriesutils import orderFileByLastLastModification
from tomwer.core.utils.locker import FileLockerManager
from tomwer.core.volume.edfvolume import EDFVolume
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.core.volume.jp2kvolume import JP2KVolume
from tomwer.core.volume.tiffvolume import TIFFVolume

logger = logging.getLogger(__name__)


class TomwerScanBase(TomwerObject):
    """
    Simple interface to extend the tomoscan.TomwerScanBase with
    specific functions
    """

    _DICT_DARK_REF_KEYS = "dark_ref_params"

    _DICT_NABU_RP_KEY = "nabu_params"

    _DICT_AXIS_KEYS = "axis_params"

    _DICT_SA_AXIS_KEYS = "sa_axis_params"

    _DICT_SA_DELTA_BETA_KEYS = "sa_delta_beta_params"

    _DICT_PROCESS_INDEX_KEY = "next_process_index"

    _DICT_NORMALIZATION_KEY = "norm_params"

    VALID_RECONS_EXTENSION = ".edf", ".npy", ".npz", ".hdf5", ".tiff", ".jp2", ".vol"

    def __init__(self, overwrite_proc_file=False):
        super().__init__()
        self._stitching_metadata = None
        self._reconstructions = []

        self._nabu_params = None
        """nabu reconstruction parameters"""
        self._axis_params = None
        """Axis parameters"""
        self._saaxis_params = None
        """Information relative to saaxis"""
        self._sa_delta_beta_params = None
        """Information regarding sa_delta_beta_params"""
        self._dark_ref_params = None
        """Information regarding dark - ref reconstruction"""
        self._process_file = None
        """file storing processes applied on the scan, with their configuration
        and result"""
        self._cache_proj_urls = None
        """cache for the projection urls"""
        self._cache_radio_axis = {}
        """cache for the radio axis. Key is tuple (mode, nearest), value is
        (url1, url2)"""
        self._notify_ffc_rsc_missing = True
        """Should we notify the user if ffc fails because cannot find dark or
        flat. Used to avoid several warnings. Only display one"""
        self._latest_reconstructions = []
        "list of url related to latest slice reconstruction from nabu"
        self._latest_vol_reconstructions = []
        """list of url related to latest volume reconstruction from nabu"""
        self._reconstruction_paths = set()
        self._proposal_name = None

    def _clear_heavy_cache(self):
        """For scan for now we don't want to remove any information from the cache.
        Mor eusefull for volume use case
        """
        pass

    @property
    def working_directory(self) -> Optional[str]:
        """
        working directory to use for this scan (for launching reconstruction for example)
        """
        raise NotImplementedError("Base class")

    def _init_index_process_file(self, overwrite_proc_file=False):
        if (
            not overwrite_proc_file
            and self.process_file is not None
            and os.path.exists(self.process_file)
        ):
            with open_hdf5(self.process_file) as h5s:
                self._process_index = len(h5s.items())
        else:
            self._process_index = 0

    def clear_caches(self):
        self._cache_proj_urls = None
        self._notify_ffc_rsc_missing = True
        super().clear_caches()

    @staticmethod
    def get_process_file_name(scan):
        raise NotImplementedError("Base class")

    def _flat_field_correction(
        self,
        data,
        index_proj: typing.Union[int, None],
        dark,
        flat1,
        flat2,
        index_flat1: int,
        index_flat2: int,
    ):
        """
        compute flat field correction for a provided data from is index
        one dark and two flats (require also indexes)
        """
        assert type(data) is numpy.ndarray
        can_process = True

        if dark is None:
            if self._notify_ffc_rsc_missing:
                logger.warning("cannot make flat field correction, dark not found")
            can_process = False

        if dark is not None and dark.ndim != 2:
            logger.error(
                "cannot make flat field correction, dark should be of " "dimension 2"
            )
            can_process = False

        if flat1 is None:
            if self._notify_ffc_rsc_missing:
                logger.warning("cannot make flat field correction, flat not found")
            can_process = False
        else:
            if flat1.ndim != 2:
                logger.error(
                    "cannot make flat field correction, flat should be of "
                    "dimension 2"
                )
                can_process = False
            if flat2 is not None and flat1.shape != flat2.shape:
                logger.error("the tow flats provided have different shapes.")
                can_process = False

        if dark is not None and flat1 is not None and dark.shape != flat1.shape:
            logger.error("Given dark and flat have incoherent dimension")
            can_process = False

        if dark is not None and data.shape != dark.shape:
            logger.error(
                "Image has invalid shape. Cannot apply flat field" "correction it"
            )
            can_process = False

        if can_process is False:
            self._notify_ffc_rsc_missing = False
            return data

        if flat2 is None:
            flat_value = flat1
        else:
            # compute weight and clip it if necessary
            if index_proj is None:
                w = 0.5
            else:
                w = (index_proj - index_flat1) / (index_flat2 - index_flat1)
                w = min(1, w)
                w = max(0, w)
            flat_value = flat1 * w + flat2 * (1 - w)

        div = flat_value - dark
        div[div == 0] = 1
        return (data - dark) / div

    def acquire_process_file_lock(self):
        """create a FileLocker context manager to insure safe write to the
        process file"""
        if self.process_file is None:
            raise ValueError("No processing file defined")
        return FileLockerManager().get_lock(file_name=self.process_file)

    @property
    def reconstructions(self):
        """list of reconstruction files"""
        return self._reconstructions

    @reconstructions.setter
    def reconstructions(self, reconstructions):
        self._reconstructions = reconstructions

    @property
    def reconstruction_paths(self):
        return self._reconstruction_paths

    def add_reconstruction_path(self, path: str):
        self._reconstruction_paths.add(path)

    @property
    def nabu_recons_params(self):
        return self._nabu_params

    @nabu_recons_params.setter
    def nabu_recons_params(self, recons_params):
        self._nabu_params = recons_params

    @property
    def axis_params(self):
        return self._axis_params

    @axis_params.setter
    def axis_params(self, parameters):
        self._axis_params = parameters

    @property
    def saaxis_params(self):
        return self._saaxis_params

    @saaxis_params.setter
    def saaxis_params(self, saaxis_params):
        self._saaxis_params = saaxis_params

    @property
    def sa_delta_beta_params(self):
        return self._sa_delta_beta_params

    @sa_delta_beta_params.setter
    def sa_delta_beta_params(self, sa_delta_beta_params):
        self._sa_delta_beta_params = sa_delta_beta_params

    # TODO: change name. Should be generalized to return Dataurl
    def getReconstructedFilesFromParFile(self, with_index):
        raise NotImplementedError("Base class")

    def projections_with_angle(self):
        raise NotImplementedError("Base class")

    def scan_dir_name(self) -> Optional[str]:
        """return name of the directory containing the acquisition"""
        raise NotImplementedError("Base class")

    def scan_basename(self) -> Optional[str]:
        """return basename of the directory containing the acquisition"""
        raise NotImplementedError("Base class")

    def scan_parent_dir_basename(self) -> Optional[str]:
        """return parent basename of the directory containing the acquisition"""
        raise NotImplementedError("Base class")

    def pop_process_index(self) -> int:
        """Return and lock the next free process index"""
        process_index = self._process_index
        self._process_index += 1
        return process_index

    @functools.lru_cache(maxsize=3)
    def get_opposite_projections(self, mode) -> tuple:
        """
        Return the radios for axis calculation and the requested mode.

        :param: angles we want to use for COR calculation. Can be 0-180,
                 90-180 or manual. If manual will return the 'most' appropriate
        :type: CorAngleMode
        :param nearest: if True then, pick the closest angle from the requested
                        one. If False, return (None, None) if the the angles
                        does not exists.
        :type: bool
        :return: couple of `opposite` radios that can be used for axis
                 calculation.
        :rtype: tuple(Optional[AxisResource], Optional[AxisResource])
        """
        radios_with_angle = self.projections_with_angle()
        angles = numpy.array(
            tuple(
                filter(
                    lambda a: numpy.issubdtype(type(a), numpy.number),
                    radios_with_angle.keys(),
                )
            )
        )

        from ..process.reconstruction.axis.anglemode import CorAngleMode
        from ..process.reconstruction.axis.params import (
            AxisResource,
        )  # avoid cyclic import

        mode = CorAngleMode.from_value(mode)

        if len(angles) < 2:
            logger.warning("less than two angles found. Unable to get opposite angles")
            return None, None

        initial_angle = angles[0]
        if mode in (CorAngleMode.use_0_180, CorAngleMode.manual_selection):
            couples = (initial_angle, initial_angle + 180.0)
        elif mode is CorAngleMode.use_90_270:
            couples = (initial_angle + 90.0, initial_angle + 270.0)
        else:
            raise ValueError(f"{mode} is not handle")

        def find_nearest(angles: numpy.ndarray, angle: float):
            if len(angles) == 0:
                return None
            dist = numpy.abs(angles - angle)
            idx = dist.argmin()
            if isinstance(idx, numpy.ndarray):
                idx = idx[0]
            return angles[idx]

        nearest_c1 = find_nearest(angles=angles, angle=couples[0])
        nearest_c2 = find_nearest(angles=angles, angle=couples[1])
        if nearest_c1 is not None and nearest_c2 is not None:
            radio_0 = AxisResource(radios_with_angle[nearest_c1])
            radio_1 = AxisResource(radios_with_angle[nearest_c2])
            return radio_0, radio_1
        else:
            return None, None

    def data_flat_field_correction(self, data, index=None):
        """Apply flat field correction on the given data

        :param numpy.ndarray data: the data to apply correction on
        :param Uion[int, None] index: index of the data in the acquisition
                                      sequence
        :return: corrected data
        :rtype: numpy.ndarray
        """
        raise NotImplementedError("Base class")

    def getReconsParamList(self):
        """

        :return: reconstruction parameters
        :rtype: ReconsParamList
        """
        raise NotImplementedError("Base class")

    @property
    def process_file(self) -> str:
        """

        :return: file used to store the processes launch by tomwer
        """
        return self._process_file

    @property
    def process_file_url(self) -> DataUrl:
        """

        :return: DataUrl of the process file
        """
        entry = self.entry if hasattr(self, "entry") else "entry"
        return DataUrl(file_path=self._process_file, data_path=entry, scheme="silx")

    def to_dict(self):
        res = {}
        # nabu reconstruction parameters
        if self._nabu_params:
            res[self._DICT_NABU_RP_KEY] = (
                self.nabu_recons_params
                if isinstance(self.nabu_recons_params, dict)
                else self.nabu_recons_params.to_dict()
            )
        else:
            res[self._DICT_NABU_RP_KEY] = None
        # axis reconstruction parameters
        if self.axis_params is None:
            res[self._DICT_AXIS_KEYS] = None
        else:
            res[self._DICT_AXIS_KEYS] = self.axis_params.to_dict()
        # saaxis reconstruction parameters
        if self.saaxis_params is None:
            res[self._DICT_SA_AXIS_KEYS] = None
        else:
            res[self._DICT_SA_AXIS_KEYS] = self.saaxis_params.to_dict()
        # sa delta-beta reconstruction parameters
        if self._sa_delta_beta_params is None:
            res[self._DICT_SA_DELTA_BETA_KEYS] = None
        else:
            res[self._DICT_SA_DELTA_BETA_KEYS] = self.sa_delta_beta_params.to_dict()
        # dark ref
        if self._dark_ref_params is None:
            res[self._DICT_DARK_REF_KEYS] = None
        else:
            res[self._DICT_DARK_REF_KEYS] = self._dark_ref_params.to_dict()
        # normalization
        if self.intensity_normalization is None:
            res[self._DICT_NORMALIZATION_KEY] = None
        else:
            res[self._DICT_NORMALIZATION_KEY] = self.intensity_normalization.to_dict()
        # process index
        res[self._DICT_PROCESS_INDEX_KEY] = self._process_index

        return res

    def load_from_dict(self, desc):
        from tomwer.core.process.reconstruction.axis.params import (
            AxisRP,
        )  # avoid cyclic import

        if isinstance(desc, io.TextIOWrapper):
            data = json.load(desc)
        else:
            data = desc
        if not (
            self.DICT_PATH_KEY in data  # pylint: disable=E1101
            and data[self.DICT_TYPE_KEY] == self._TYPE  # pylint: disable=E1101
        ):
            raise ValueError("Description is not an EDFScan json description")

        assert self.DICT_PATH_KEY in data  # pylint: disable=E1101
        # load axis reconstruction parameters
        axis_params = data.get(self._DICT_AXIS_KEYS, None)
        if axis_params is not None:
            self.axis_params = AxisRP.from_dict(axis_params)
        # load nabu reconstruction parameters
        if self._DICT_NABU_RP_KEY in data:
            self._nabu_params = data[self._DICT_NABU_RP_KEY]
        # load dark-ref parameters
        dark_ref_params = data.get(self._DICT_DARK_REF_KEYS, None)
        if dark_ref_params is not None:
            from tomwer.core.process.reconstruction.darkref.params import DKRFRP

            self._dark_ref_params = DKRFRP.from_dict(dark_ref_params)
        # load normalization
        intensity_normalization = data.get(self._DICT_NORMALIZATION_KEY, None)
        if intensity_normalization is not None:
            self.intensity_normalization = IntensityNormalization.from_dict(
                intensity_normalization
            )
        # load saaxis parameters
        saaxis_params = data.get(self._DICT_SA_AXIS_KEYS, None)
        if saaxis_params is not None:
            from tomwer.core.process.reconstruction.saaxis.params import SAAxisParams

            self._saaxis_params = SAAxisParams.from_dict(saaxis_params)
        # load sa delta beta parameters
        sa_delta_beta_params = data.get(self._DICT_SA_DELTA_BETA_KEYS, None)
        if sa_delta_beta_params is not None:
            from tomwer.core.process.reconstruction.sadeltabeta.params import (
                SADeltaBetaParams,
            )

            self._sa_delta_beta_params = SADeltaBetaParams.from_dict(
                sa_delta_beta_params
            )

        self._process_index = data[self._DICT_PROCESS_INDEX_KEY]

    def equal(self, other):
        """

        :param :class:`.ScanBase` other: instance to compare with
        :return: True if instance are equivalent
        :note: we cannot use the __eq__ function because this object need to be
               pickable
        """
        return (
            isinstance(other, self.__class__)
            or isinstance(self, other.__class__)
            and self.type == other.type  # pylint: disable=E1101
            and self.nabu_recons_params == other.nabu_recons_params
            and self.path == other.path  # pylint: disable=E1101
        )

    def get_sinogram(self, line, subsampling=1, norm_method=None, **kwargs):
        """
        extract the sinogram from projections

        :param line: which sinogram we want
        :type: int
        :param subsampling: subsampling to apply on the sinogram
        :return: sinogram from the radio lines
        :rtype: numpy.array
        """
        raise NotImplementedError("Base class")

    def get_normed_sinogram(self, line, subsampling=1):
        """
        Util to get the sinogram normed with settings currently defined
        on the 'intensity_normalization' property

        :param line:
        :param subsampling:
        :return:
        """
        return self.get_sinogram(
            line=line,
            subsampling=subsampling,
            norm_method=self.intensity_normalization.method,
            **self.intensity_normalization.get_extra_infos(),
        )

    def __str__(self):
        raise NotImplementedError("Base class")

    @staticmethod
    def get_pyhst_recons_file(scanID):
        """Return the .par file used for the current reconstruction if any.
        Otherwise return None"""
        if scanID == "":
            return None

        if scanID is None:
            raise RuntimeError("No current acquisition to validate")
        assert type(scanID) is str
        assert os.path.isdir(scanID)
        folderID = os.path.basename(scanID)
        # look for fasttomo files ending by slice.par
        parFiles = glob(os.path.join(scanID + folderID) + "*_slice.par")
        if len(parFiles) > 0:
            return orderFileByLastLastModification(scanID, parFiles)[-1]
        else:
            return None

    def _deduce_transfert_scan(self, output_dir):
        """
        Create the scan that will be generated if this scan is
        copy to the output_dir

        :param str output_dir:
        """
        raise NotImplementedError("Base class")

    def get_proj_angle_url(self, use_cache: bool = True, *args, **kwargs):
        """
        retrieve the url for each projections (including the alignement /
        return one) and associate to each (if possible) the angle.
        Alignment angle are set as angle (1) to specify that this is an
        alignment one.
        :param bool use_cache:
        :return: dictionary with angle (str or int) as key and url as value
        :rtype: dict
        """
        raise NotImplementedError("Base class")

    def get_reconstructions_urls(self):
        """

        :return: list of urls that contains reconstruction from nabu
        :rtype: list
        """
        from tomwer.core.process.reconstruction.output import (
            PROCESS_FOLDER_NAME,
        )  # avoid cyclic import

        all_recons_urls = set()
        recons_paths = self.reconstruction_paths
        if self.working_directory:
            recons_paths.add(
                os.path.join(
                    self.working_directory,
                    PROCESS_FOLDER_NAME,
                )
            )
        for path in recons_paths:
            all_recons_urls.update(self.get_reconstructions_urls_from_path(path))
        return all_recons_urls

    def get_reconstructions_urls_from_path(self, path, check_url=False):
        if path is None or not os.path.isdir(path):
            return []
        else:

            from tomwer.core.process.reconstruction.saaxis.saaxis import (
                DEFAULT_RECONS_FOLDER as MULTI_COR_DEFAULT_FOLDER,
            )
            from tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta import (
                DEFAULT_RECONS_FOLDER as MULTI_DB_DEFAULT_FOLDER,
            )
            from tomwer.core.process.reconstruction.nabu.settings import (
                NABU_CFG_FILE_FOLDER,
            )

            res = set()
            for f in os.listdir(path):
                current_path = os.path.join(path, f)
                if os.path.isfile(current_path):
                    volume = _get_reconstructed_single_file_volume(
                        current_path, scan=self
                    )
                    if volume is not None:
                        res.add(volume)
                # avoid looking deeper if this doesn't look like a folder that could contain slices
                elif (
                    f
                    in (
                        "matlab",
                        "steps_file_basename_nabu_sinogram_save_step",
                        MULTI_COR_DEFAULT_FOLDER,
                        MULTI_DB_DEFAULT_FOLDER,
                        NABU_CFG_FILE_FOLDER,
                    )
                    or f.endswith("_vol")
                    or "slice" not in f
                ):
                    pass
                else:
                    res.update(
                        self.get_reconstructions_urls_from_path(
                            path=current_path, check_url=check_url
                        )
                    )
            return res

    @property
    def latest_reconstructions(self):
        """List of latest slices reconstructions (as VolumeIdentifier) - single slice volume"""
        return self._latest_reconstructions

    @property
    def latest_vol_reconstructions(self):
        """List of latest volume reconstructions (as VolumeIdentifier)"""
        return self._latest_vol_reconstructions

    def clear_latest_reconstructions(self):
        self._latest_reconstructions = []

    def set_latest_reconstructions(self, urls: typing.Iterable):
        if urls is None:
            self._latest_reconstructions = None
        else:
            self._latest_reconstructions = list(
                [self._process_volume_url(url) for url in urls]
            )

    def add_latest_reconstructions(self, urls: typing.Iterable):
        self._latest_reconstructions.extend(urls)

    def clear_latest_vol_reconstructions(self):
        self._latest_vol_reconstructions = []

    @staticmethod
    def _process_volume_url(url):
        if isinstance(url, str):
            return VolumeIdentifier.from_str(url)
        elif isinstance(url, VolumeIdentifier):
            return url
        elif isinstance(url, VolumeBase):
            return url.get_identifier()
        else:
            raise TypeError(
                f"url should be a {VolumeIdentifier} or a string reprenseting a {VolumeIdentifier}. {type(url)} provided instead"
            )

    def set_latest_vol_reconstructions(self, volume_identifiers: typing.Iterable):
        if volume_identifiers is None:
            self._latest_vol_reconstructions = None
        else:
            self._latest_vol_reconstructions = list(
                [self._process_volume_url(url) for url in volume_identifiers]
            )

    def add_latest_vol_reconstructions(self, volume_identifiers: tuple):
        assert isinstance(
            volume_identifiers, tuple
        ), "volume_identifiers is expected to be a tuple"
        self._latest_vol_reconstructions.extend(
            self._process_volume_url(volume_identifier)
            for volume_identifier in volume_identifiers
        )

    def _update_latest_recons_identifiers(self, old_path, new_path):
        def update_identifier(identifier):
            assert isinstance(
                identifier, BaseIdentifier
            ), f"identifier is expected to be an instance of {BaseIdentifier}"
            # small hack as this is not much used: simply try to replace a path by another. this is only used by the data transfer and EDF / SPEC
            # this time is almost over
            # FIXME: correct way to do this would be to recreate the volume, modify file or folder path and
            # recreate the new identifier
            identifier.replace(old_path, new_path, 1)

        self._latest_reconstructions = [
            update_identifier(identifier=identifier)
            for identifier in self._latest_reconstructions
        ]

        self._latest_vol_reconstructions = [
            update_identifier(identifier=identifier)
            for identifier in self._latest_vol_reconstructions
        ]

    def get_url_proj_index(self, url):
        """Return the index in the acquisition from the url"""

        def invert_dict(ddict):
            res = {}
            if ddict is not None:
                for key, value in ddict.items():
                    assert isinstance(value, DataUrl)
                    res[value.path()] = key
            return res

        proj_inv_url_to_index = invert_dict(self.projections)  # pylint: disable=E1101
        alig_inv_url_to_index = invert_dict(
            self.alignment_projections  # pylint: disable=E1101
        )
        if url.path() in proj_inv_url_to_index:
            return proj_inv_url_to_index[url.path()]
        elif url.path() in alig_inv_url_to_index:
            return alig_inv_url_to_index[url.path()]
        else:
            return None

    def set_process_index_frm_tomwer_process_file(self):
        """
        Set the process_index to the last index find in the tomwer_process_file
        + 1
        """
        from tomwer.core.process.task import Task

        if os.path.exists(self.process_file):
            with open_hdf5(self.process_file) as h5s:
                if not hasattr(self, "entry"):
                    entry = "entry"
                else:
                    entry = self.entry
                if entry in h5s:
                    node = h5s[entry]
                    pn = Task._get_process_nodes(
                        root_node=node, process=None, version=None
                    )
                    indexes = pn.values()
                    if len(indexes) > 0:
                        self._process_index = max(indexes) + 1
                        logger.debug(
                            f"set process_index from tomwer process file to {self._process_index}"
                        )

    def get_nabu_dataset_info(self, binning=1, binning_z=1, proj_subsampling=1):
        """

        :return: nabu dataset descriptor
        :rtype: dict
        """
        raise NotImplementedError("Base class")

    def to_nabu_dataset_analyser(self):
        """Return the equivalent DatasetAnalyzer for nabu"""
        raise NotImplementedError("Base class")

    def get_proposal_name(self) -> Optional[str]:
        return self._proposal_name

    def set_proposal_name(self, proposal_name: str) -> None:
        self._proposal_name = proposal_name


class _TomwerBaseDock(object):
    """
    Internal class to make difference between a simple TomoBase output and
    an output for a different processing (like scanvalidator.UpdateReconsParam)
    """

    def __init__(self, tomo_instance):
        self.__instance = tomo_instance

    @property
    def instance(self):
        return self.__instance


def _containsDigits(input_):
    return any(char.isdigit() for char in input_)


def _get_reconstructed_single_file_volume(
    file_, scan: TomwerScanBase, check_url=False
) -> Optional[VolumeBase]:
    scan_basename = scan.get_process_file_name
    scan_basename = scan.get_dataset_basename()
    if scan_basename is None:
        logger.error("TODO")

    file_base_name = os.path.basename(file_)
    if file_base_name.startswith(scan_basename) and "slice_" in file_base_name:
        if file_.endswith(TomwerScanBase.VALID_RECONS_EXTENSION):
            local_str = file_

            for extension in TomwerScanBase.VALID_RECONS_EXTENSION:
                if local_str.endswith(extension):
                    local_str = local_str.rsplit(extension, 1)[0]
                    break
            if "slice_pag_" in local_str:
                indexStr = local_str.split("slice_pag_")[-1].split("_")[0]
                volume_basename = local_str.split("slice_pag_")[0] + "slice_pag_"
            else:
                indexStr = local_str.split("slice_")[-1].split("_")[0]
                volume_basename = local_str.split("slice_")[0] + "slice_"

            if _containsDigits(indexStr):
                if extension == ".edf":
                    return EDFVolume(
                        folder=os.path.dirname(file_),
                        volume_basename=volume_basename,
                    )

            if extension in (".jp2", ".jp2"):
                return JP2KVolume(
                    folder=os.path.dirname(file_),
                    volume_basename=volume_basename,
                    data_extension=extension,
                )
            elif extension in (".tiff", ".tif"):
                return TIFFVolume(
                    folder=os.path.dirname(file_),
                    volume_basename=volume_basename,
                    data_extension=extension,
                )
            elif extension in (
                (".h5", ".hdf5", ".nx", ".nxs", ".nexus", ".nx5", ".hdf")
            ):
                entry = getattr(scan, "entry", "entry")
                volume = HDF5Volume(
                    file_path=os.path.abspath(file_),
                    data_path="/".join([entry, "reconstruction"]),
                )

                if check_url is True:
                    try:
                        with open_hdf5(os.path.abspath(file_)) as h5f:
                            if entry not in h5f:
                                logger.info("{volume} does not exists")
                                return None
                    except Exception:
                        logger.info("unable to check {volume}")
                        return None
                    else:
                        logger.info("{volume} checked")
                return volume


class NormMethod(_Enum):
    MANUAL_ROI = "manual ROI"
    AUTO_ROI = "automatic ROI"
    DATASET = "from dataset"



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "03/02/2021"


import os
import shutil
import tempfile

import h5py
import numpy
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from silx.io.url import DataUrl
from pyunitsystem import metricsystem

from tomwer.core.process.reconstruction.saaxis.params import (
    ReconstructionMode,
    SAAxisParams,
)
from tomwer.core.process.reconstruction.scores import ComputedScore
from tomwer.core.process.reconstruction.scores.params import ScoreMethod
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.gui.reconstruction.saaxis.dimensionwidget import DimensionWidget
from tomwer.gui.reconstruction.saaxis.saaxis import SAAxisWindow
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestDimensionWidget(TestCaseQt):
    """Test that the axis widget work correctly"""

    def setUp(self):
        self._window = DimensionWidget(title="test window")

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()

    def testDimension(self):
        """Insure setting dimensions values and unit are correct"""
        self.assertTrue(self._window.unit() is metricsystem.MetricSystem.MILLIMETER)
        # check initial values
        self.assertEqual(
            self._window.getValues(),
            (0.001, 0.001, 0.001, metricsystem.MetricSystem.METER),
        )
        # set dim 0 to 20 mm
        self._window._dim0ValueQLE.setValue(20)
        self._window._dim0ValueQLE.editingFinished.emit()
        self.qapp.processEvents()
        self.assertEqual(
            self._window.getValues(),
            (0.02, 0.001, 0.001, metricsystem.MetricSystem.METER),
        )
        # set dim 1 to 0.6 millimeter
        self._window._dim1ValueQLE.setValue(0.6)
        self._window._dim1ValueQLE.editingFinished.emit()
        self.qapp.processEvents()
        self.assertEqual(
            self._window.getValues(),
            (0.02, 0.0006, 0.001, metricsystem.MetricSystem.METER),
        )
        # change display to micrometer
        self._window.setUnit(metricsystem.MetricSystem.MICROMETER)
        self.assertEqual(
            self._window.getValues(),
            (0.02, 0.0006, 0.001, metricsystem.MetricSystem.METER),
        )
        self.assertEqual(
            self._window._dim0ValueQLE.value(),
            20
            * metricsystem.MetricSystem.MILLIMETER.value
            / metricsystem.MetricSystem.MICROMETER.value,
        )
        self._window._dim2ValueQLE.setValue(500)
        self._window._dim2ValueQLE.editingFinished.emit()
        self.qapp.processEvents()
        self.assertEqual(
            self._window.getValues(),
            (0.02, 0.0006, 0.0005, metricsystem.MetricSystem.METER),
        )

    def testConstructorWithColors(self):
        """Insure passing colors works"""
        DimensionWidget(title="title", dims_colors=("#ffff5a", "#62efff", "#ff5bff"))

    def testConstructorWithdimNames(self):
        """Insure passing colors works"""
        DimensionWidget(title="title", dims_name=("x", "y", "z"))


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestSAAxisWindow(TestCaseQt):
    """Test the SAAxisWindow interface"""

    _N_OUTPUT_URLS = 10

    def setUp(self):
        super().setUp()
        self._window = SAAxisWindow()
        self.folder = tempfile.mkdtemp()
        self._output_urls = []
        self._cor_scores = {}

        for i in range(self._N_OUTPUT_URLS):
            output_file = os.path.join(self.folder, f"recons_{i}.h5")
            with h5py.File(output_file, mode="a") as h5f:
                h5f["data"] = numpy.random.random(100 * 100).reshape(100, 100)
                url = DataUrl(file_path=output_file, data_path="data", scheme="silx")
                assert url.is_valid()
            self._output_urls.append(url)
            score = ComputedScore(
                tv=numpy.random.randint(10),
                std=numpy.random.randint(100),
            )
            self._cor_scores[i] = (url, score)

        # create a scan
        self.folder = tempfile.mkdtemp()
        dim = 10
        mock = MockNXtomo(
            scan_path=self.folder, n_proj=10, n_ini_proj=10, scan_range=180, dim=dim
        )
        mock.add_alignment_radio(index=10, angle=90)
        mock.add_alignment_radio(index=10, angle=0)
        self.scan = mock.scan

        self._window.setScan(self.scan)

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        self._window = None
        shutil.rmtree(self.folder)
        super().tearDown()

    def testGetConfiguration(self):
        """Insure configuration getter works"""
        conf = self._window.getConfiguration()
        # the configuration should be convertible to
        conf = SAAxisParams.from_dict(conf)
        self.assertEqual(conf.slice_indexes, "middle")
        self.assertEqual(conf.file_format, "hdf5")
        self.assertTrue(isinstance(conf.nabu_params, dict))
        for key in (
            "preproc",
            "reconstruction",
            "dataset",
            "tomwer_slices",
            "output",
            "phase",
        ):
            with self.subTest(key=key):
                self.assertTrue(key in conf.nabu_params)
        self.assertEqual(conf._dry_run, False)
        self.assertEqual(conf.mode, ReconstructionMode.VERTICAL)
        self.assertEqual(conf.score_method, ScoreMethod.TV_INVERSE)
        self.assertEqual(conf.output_dir, None)
        self.assertEqual(conf.scores, None)

    def testSetConfiguration(self):
        """Insure configuration setter works"""
        self._window.setSlicesRange(0, 20)
        configuration = {
            "slice_index": {"Slice": 2},
            "estimated_cor": 0.24,
            "research_width": 0.8,
            "n_reconstruction": 5,
        }
        self._window.setConfiguration(configuration)
        res_conf = self._window.getConfiguration()
        for key in (
            "slice_index",
            "estimated_cor",
            "research_width",
            "n_reconstruction",
        ):
            self.assertEqual(configuration[key], res_conf[key])

    def testSetResults(self):
        """Test setting results and saving result to a folder"""
        self._window.setCorScores(self._cor_scores, score_method="standard deviation")
        # test saving snapshots
        with tempfile.TemporaryDirectory() as output_png_imgs:
            final_dir = os.path.join(output_png_imgs, "test/create/it")
            self._window.saveReconstructedSlicesTo(final_dir)
            assert os.path.exists(final_dir)
            assert len(os.listdir(final_dir)) == len(self._cor_scores)



import logging
import os
from math import floor

import numpy
from nxtomomill.converter.hdf5.utils import PROCESSED_DATA_DIR_NAME, RAW_DATA_DIR_NAME
from PIL import Image
from silx.io.url import DataUrl
from silx.utils.enum import Enum as _Enum
from tomoscan.esrf.scan.utils import get_data
from tomoscan.esrf.volume.hdf5volume import HDF5Volume
from tomoscan.esrf.volume.singleframebase import VolumeSingleFrameBase

from tomwer.io.utils.raw_and_processed_data import to_processed_data_path
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.process.task import Task
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.process.icat.screenshots import IcatScreenshots
from tomwer.core.process.icat.publish import (
    PublishProcessedDataFolderTask,
    from_bliss_original_file_to_raw,
)
from processview.core.manager import DatasetState, ProcessManager

_logger = logging.getLogger(__name__)

PROPOSAL_GALLERY_DIR_NAME = "GALLERY"
DATASET_GALLERY_DIR_NAME = "gallery"


class OutputFormat(_Enum):
    """possible output format to save screenshots"""

    PNG = "png"
    JPEG = "jpg"


class Binning(_Enum):
    ONE_BY_ONE = "1x1"
    TWO_BY_TWO = "2x2"
    FOUR_BY_FOUR = "4x4"
    HEIGHT_BY_HEIGHT = "8x8"
    SIXTEEN_BY_SIXTEEN = "16x16"
    THIRTY_TWO_BY_THIRTY_TWO = "32x32"
    SIXTY_FOUR_BY_SIXTY_FOUR = "64x64"
    ONE_HUNDRED_TWENTY_HEIGHT_BY_ONE_HUNDRED_TWENTY_HEIGHT = "128x128"

    @staticmethod
    def _bin_data(data, binning):
        if not isinstance(data, numpy.ndarray):
            raise TypeError("data should be a numpy array")
        if not data.ndim == 2:
            raise ValueError("data is expected to be 2d")
        binning = Binning.from_value(binning)
        if binning is Binning.ONE_BY_ONE:
            return data
        elif binning is Binning.TWO_BY_TWO:
            return data[::2, ::2]
        elif binning is Binning.FOUR_BY_FOUR:
            return data[::4, ::4]
        elif binning is Binning.HEIGHT_BY_HEIGHT:
            return data[::8, ::8]
        elif binning is Binning.SIXTEEN_BY_SIXTEEN:
            return data[::16, ::16]
        elif binning is Binning.THIRTY_TWO_BY_THIRTY_TWO:
            return data[::32, ::32]
        elif binning is Binning.SIXTY_FOUR_BY_SIXTY_FOUR:
            return data[::64, ::64]
        else:
            raise NotImplementedError


def deduce_dataset_gallery_location(scan_obj: TomwerScanBase) -> str:
    """
    From scan path deduce the 'dataset' path to the gallery.
    Warning: dataset gallery is different then the 'proposal' gallery
    """
    if not isinstance(scan_obj, TomwerScanBase):
        raise TypeError(f"'scan_obj' is expected to be an instance of {TomwerScanBase}")

    file_path = os.path.abspath(scan_obj.path)

    split_path = file_path.split(os.sep)
    # reverse it to find the lower level value of 'PROCESSED_DATA_DIR_NAME' or 'RAW_DATA_DIR_NAME' if by any 'chance' has several in the path
    # then we will replace the 'lower one' in the string. This is where the GALLERY will be added
    split_path = split_path[::-1]
    # check if already contain in a "PROCESSED_DATA" directory
    try:
        index_processed_data = split_path.index(PROCESSED_DATA_DIR_NAME)
    except ValueError:
        pass
        index_processed_data = None
    try:
        index_raw_data = split_path.index(RAW_DATA_DIR_NAME)
    except ValueError:
        # if the value is not in the list
        index_raw_data = None

    if index_processed_data is None and index_raw_data is None:
        # if not in any "PROCESSED_DATA" or "RAW_DATA" directory
        return scan_obj.get_relative_file(
            file_name=DATASET_GALLERY_DIR_NAME, with_dataset_prefix=False
        )
    elif index_processed_data is not None and index_raw_data is not None:
        if index_raw_data > index_processed_data:
            # if PROCESSED_DATA lower in the path than RAW_DATA
            split_path[index_processed_data] = RAW_DATA_DIR_NAME
    # reorder path to original
    split_path = list(split_path[::-1])
    split_path.append(DATASET_GALLERY_DIR_NAME)
    # move it to PPROCESSED_DATA when possible
    path = os.sep.join(split_path)
    path = to_processed_data_path(path)
    return path


def deduce_proposal_GALLERY_location(scan_obj: TomwerScanBase) -> str:
    """
    Policy: look if the scan_obj.path is in 'PROCESSED_DATA_DIR_NAME' or 'RAW_DATA_DIR_NAME' directories.
    If find any (before any 'GALLERY_DIR_NAME' directory) replace it "GALLERY_DIR_NAME".
    If none of those are found then create it at the same level as the scan

    :param TomwerScanBase scan_obj: scan_obj for which we want the GALLERY directory
    :return: gallery path (to save screeshots for example)
    """
    if not isinstance(scan_obj, TomwerScanBase):
        raise TypeError(f"'scan_obj' is expected to be an instance of {TomwerScanBase}")

    file_path = os.path.abspath(scan_obj.path)

    split_path = file_path.split(os.sep)
    # reverse it to find the lower level value of 'PROCESSED_DATA_DIR_NAME' or 'RAW_DATA_DIR_NAME' if by any 'chance' has several in the path
    # then we will replace the 'lower one' in the string. This is where the GALLERY will be added
    split_path = split_path[::-1]
    # check if already contain in a "PROCESSED_DATA" directory
    try:
        index_processed_data = split_path.index(PROCESSED_DATA_DIR_NAME)
    except ValueError:
        pass
        index_processed_data = None
    try:
        index_raw_data = split_path.index(RAW_DATA_DIR_NAME)
    except ValueError:
        # if the value is not in the list
        index_raw_data = None

    if index_processed_data is None and index_raw_data is None:
        # if not in any "PROCESSED_DATA" or "RAW_DATA" directory
        return scan_obj.get_relative_file(
            file_name=PROPOSAL_GALLERY_DIR_NAME, with_dataset_prefix=False
        )
    elif index_processed_data is not None and index_raw_data is not None:
        if index_raw_data > index_processed_data:
            # if PROCESSED_DATA lower in the path than RAW_DATA
            split_path[index_processed_data] = PROPOSAL_GALLERY_DIR_NAME
        else:
            # if RAW_DATA lower in the path than PROCESSED_DATA
            split_path[index_raw_data] = PROPOSAL_GALLERY_DIR_NAME
    elif index_raw_data is not None:
        # if the path contains only PROCESSED_DATA or RAW_DATA (expected behavior for online acquistion)
        split_path[index_raw_data] = PROPOSAL_GALLERY_DIR_NAME
    else:
        assert index_processed_data is not None, "index_processed_data is None"
        split_path[index_processed_data] = PROPOSAL_GALLERY_DIR_NAME

    # reorder path to original
    split_path = split_path[::-1]
    return os.sep.join(split_path)


def select_screenshot_from_volume(volume: TomwerVolumeBase) -> dict:
    """
    return a subset of url for a volume.
    Warning: this function will be called each time a nabu slice or a nabu volume is executer from orangecontrig.
    So it must stay 'low processing' function to avoid slowing down everything
    """
    if not isinstance(volume, TomwerVolumeBase):
        raise TypeError(
            f"volume is expected to be an instance of {TomwerVolumeBase}. Get {type(volume)}"
        )

    if isinstance(volume, VolumeSingleFrameBase):
        volume_urls = tuple(volume.browse_data_urls())
        # lets take the three equally spaced slices (3/6, 4/6 and 5/6)
        n_slices = len(volume_urls)
        # note: using a dict ensure to get a set of DataUrl in case an Url is used several time
        # workaround: DataUrl is not hashable
        screenshots = {
            f"{os.path.splitext(os.path.basename(volume_urls[floor(i * n_slices / 6)].file_path()))}": volume_urls[
                floor(i * n_slices / 6)
            ]
            for i in range(2, 5)
        }
        return screenshots
    elif isinstance(volume, HDF5Volume):
        volume_url = next(volume.browse_data_urls())
        n_slices = volume.get_volume_shape()[0]
        # note: using a dict ensure to get a set of DataUrl in case an Url is used several time
        # workaround: DataUrl is not hashable
        screenshots = {
            f"{os.path.splitext(os.path.basename(volume_url.file_path()))[0]}_{floor(i / 6 * n_slices)}": DataUrl(
                file_path=volume_url.file_path(),
                data_path=volume_url.data_path(),
                scheme="silx",
                data_slice=floor(i / 6 * n_slices),
            )
            for i in range(2, 5)
        }
        return screenshots
    else:
        _logger.warning(
            f"volume {type(volume)} does not allow to create screenshot for now"
        )
        return ()


class SaveScreenshotsToGalleryTask(
    Task,
    input_names=("screenshots",),  # screenshots as instance of :class:Screenshots
    optional_input_names=("format", "overwrite", "binning"),
):
    """simple task to do the binding between orange design and 'SaveScreenshotsTask'"""

    def run(self):
        if not isinstance(self.inputs.screenshots, IcatScreenshots):
            raise TypeError(
                f"'screenshots' is expected to be an instance of {IcatScreenshots}. get {type(self.inputs.screenshots)}"
            )

        inputs = self.inputs.screenshots.to_dict()
        if self.get_input_value("overwrite", None) is not None:
            inputs["overwrite"] = self.inputs.overwrite
        if self.get_input_value("format", None) is not None:
            inputs["format"] = self.inputs.format
        if self.get_input_value("binning", None) is not None:
            inputs["binning"] = self.inputs.binning
        task = SaveScreenshotsTask(
            inputs=inputs,
        )
        task.run()


class SaveScreenshotsTask(
    Task,
    input_names=("screenshots_as_dict", "output_dir"),
    optional_input_names=("format", "overwrite", "binning"),
):
    """Task which save a set of screenshot to the output_dir under required format"""

    def run(self):
        format = OutputFormat.from_value(
            self.get_input_value("format", OutputFormat.PNG)
        )
        overwrite = self.get_input_value("overwrite", False)
        binning = self.get_input_value("binning", Binning.ONE_BY_ONE)

        os.makedirs(self.inputs.output_dir, exist_ok=True)
        for screenshot_name, data_url in self.inputs.screenshots_as_dict.items():
            if not isinstance(data_url, DataUrl):
                raise TypeError("screenshot values are expected to DataUrl")
            data = get_data(data_url)
            if not isinstance(data, numpy.ndarray) and data.ndim != 2:
                raise TypeError("screenshot are expected to be 2D numpy arrays")
            if not isinstance(screenshot_name, str):
                raise TypeError(
                    f"screenshot keys are expected to be str. Get {type(screenshot_name)}"
                )

            if data.ndim == 3 and data.shape[0] == 1:
                data = data.reshape(data.shape[1:])
            elif data.ndim != 2:
                raise ValueError(
                    f"only 2D grayscale image are handled. Get {data.shape}"
                )
            # if qt is available use it as this is simpler and more powerful
            # clamp data in 0-256
            data = data.astype(numpy.float32)
            data = Binning._bin_data(data=data, binning=binning)
            data *= 255.0 / data.max()

            # do a rescale else use qt instead
            img = Image.fromarray(data, mode=None)
            img = img.convert("L")
            output_file = os.path.join(
                self.inputs.output_dir,
                ".".join([screenshot_name, format.value]),
            )
            if not overwrite and os.path.exists(output_file):
                raise OSError(f"File already exists ({output_file})")
            img.save(output_file)


class SaveToGalleryAndPublishTask(
    Task,
    input_names=(
        "screenshots",
        "beamline",
        "dataset",
        "proposal",
        "format",
    ),
    optional_input_names=(
        "dry_run",
        "__process__",
    ),
):
    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        super().__init__(varinfo, inputs, node_id, node_attrs, execinfo)
        self._process = self.get_input_value("__process__", None)
        self._task_save = SaveScreenshotsToGalleryTask(
            inputs=inputs,
        )

        scan = self.inputs.screenshots.scan
        if isinstance(scan, EDFTomoScan):
            raise TypeError(
                "scan is an EDFTomoScan. Not handled for publishing processed data to icat"
            )
        raw = from_bliss_original_file_to_raw(scan.get_bliss_orginal_files())

        path = self.inputs.screenshots.data_dir
        # path = os.path.dirname(self.inputs.screenshots.data_dir)

        self._task_publish = PublishProcessedDataFolderTask(
            inputs={
                "beamline": self.inputs.beamline,
                "dataset": self.inputs.dataset,
                "proposal": self.inputs.proposal,
                "path": path,  # must be the dataset path
                "raw": raw,
                "dry_run": self.get_input_value("dry_run", False),
                "metadata": self.inputs.screenshots.metadata,
            },
        )

    def run(self):
        if self._process is not None:
            ProcessManager().notify_dataset_state(
                dataset=self.inputs.screenshots.scan,
                process=self._process(),
                state=DatasetState.ON_GOING,
            )
        # save screenshots to gallery
        try:
            self._task_save.run()
        except Exception as e:
            if self._process is not None:
                ProcessManager().notify_dataset_state(
                    dataset=self.inputs.screenshots.scan,
                    process=self._process(),
                    state=DatasetState.FAILED,
                    details=str(e),
                )
            raise e

        # publication
        try:
            self._task_publish.run()
        except Exception as e:
            if self._process is not None:
                ProcessManager().notify_dataset_state(
                    dataset=self.inputs.screenshots.scan,
                    process=self._process(),
                    state=DatasetState.FAILED,
                    details=str(e),
                )
            raise e
        else:
            if self._process is not None:
                ProcessManager().notify_dataset_state(
                    dataset=self.inputs.screenshots.scan,
                    process=self._process(),
                    state=DatasetState.SUCCEED,
                )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "30/10/2020"


import logging
from typing import Union

import h5py
from silx.gui import qt
from silx.gui.dialog.DataFileDialog import DataFileDialog
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset
from silx.io.utils import open as open_hdf5
from nxtomo.nxobject.nxdetector import ImageKey

import tomwer.core.utils.nxtomoutils as nxtomo_utils
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.io.utils import get_default_directory

_logger = logging.getLogger(__name__)


class _DarkOrFlatUrl(qt.QWidget):
    editingFinished = qt.Signal()
    """Signal emit when the url changes"""

    def __init__(self, parent, type_: ImageKey, when: str):
        assert when in ("start", "end")
        qt.QWidget.__init__(self, parent)
        self._type = type_
        self._when = when
        self.setLayout(qt.QHBoxLayout())

        self._qle = qt.QLineEdit("", self)
        self._qle.setPlaceholderText("scheme:file_path?data_path")
        self.layout().addWidget(self._qle)

        self._optionsCB = qt.QComboBox(self)
        self._optionsCB.addItem("full dataset")
        self.layout().addWidget(self._optionsCB)

        # connect signal / slot
        self._qle.editingFinished.connect(self._tryUpdateOptions)
        self._qle.editingFinished.connect(self.editingFinished)
        self._optionsCB.currentIndexChanged.connect(self.editingFinished)

    def setUrl(self, url):
        """
        Define the url where the dataset can be picked
        If this is a dataset from an NXEntry containing dark / flat then we
        will propose to pick only a part of the dataset (only start / end of
         dark / flat)

        :param url:
        :return:
        """
        url = self._redirectDataPath(url, logger=_logger)

        def dataset_invalid(url):
            with open_hdf5(url.file_path()) as h5s:
                if not isinstance(h5s[url.data_path()], h5py.Dataset):
                    return True
            return False

        if url.data_path() in (None, "") or dataset_invalid(url):
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Warning)
            text = "You should provide a path to a dataset or to a NXtomo entry"
            msg.setText(text)
            msg.exec_()
            return
        old = self.blockSignals(True)
        self._qle.setText(url.path())
        self._updateOptions(url)
        self.blockSignals(old)
        self.editingFinished.emit()

    def setSerie(self, serie):
        type_name = self._type.name.lower().replace("_", " ")
        item_txt = f"{type_name} serie n°{serie}"
        idx = self._optionsCB.findText(item_txt)
        if idx >= 0:
            self._optionsCB.setCurrentIndex(idx)

    def getSerieIndex(self):
        return self._optionsCB.currentIndex()

    def _redirectDataPath(self, url, logger=None):
        try:
            with open_hdf5(url.file_path()) as h5s:
                node = h5s[url.data_path()]

                if NXtomoScan.entry_is_nx_tomo(node):
                    if (
                        "detector" in node["instrument"]
                        and "data" in node["instrument"]["detector"]
                    ):
                        if logger:
                            _logger.info(
                                "NXTomo entry found. Set the directly the"
                                "detector data"
                            )
                        data_path = "/".join(
                            (url.data_path(), "instrument", "detector", "data")
                        )
                        url = DataUrl(file_path=url.file_path(), data_path=data_path)
                elif NXtomoScan.is_nxdetector(node):
                    if "data" in node:
                        if logger:
                            _logger.info(
                                "NX_detector entry found. Set the directly"
                                " the detector data"
                            )
                        data_path = "/".join((url.data_path, "data"))
                        url = DataUrl(file_path=url.file_path(), data_path=data_path)
        except Exception:
            pass
        return url

    def _tryUpdateOptions(self):
        try:
            url = DataUrl(path=self._qle.text())
            self._updateOptions(url)
        except Exception:
            pass

    def _updateOptions(self, url: DataUrl):
        self._optionsCB.clear()
        if url is None:
            return
        assert isinstance(url, DataUrl)
        assert h5py.is_hdf5(url.file_path()), "only manage hdf5 file"
        self._optionsCB.addItem("full dataset")
        if url.data_path() == "":
            _logger.error("data path should be specify")
            return
        # if we are on a 'detector / data dataset' then we can try to reach
        # image_key information
        image_keys = self._getImageKey(url)
        if image_keys is not None:
            n_serie = nxtomo_utils.get_n_series(
                image_key_values=image_keys, image_key_type=self._type
            )
            if n_serie is not None:
                type_name = self._type.name.lower().replace("_", " ")
                series_indexes = list(range(n_serie))
                for i_serie in series_indexes:
                    self._optionsCB.addItem(f"{type_name} serie n°{i_serie}")
                # if this is a end url then set to the last found entry
                if self._when == "end":
                    opt_idx = self._optionsCB.findText(
                        f"{type_name} serie n°{series_indexes[-1]}"
                    )
                else:
                    # else set it to the first entry
                    opt_idx = self._optionsCB.findText(
                        f"{type_name} serie n°{series_indexes[0]}"
                    )
                if opt_idx >= 0:
                    self._optionsCB.setCurrentIndex(opt_idx)

    def _getImageKey(self, url):
        # if we are on a 'detector / data dataset' then we can try to reach
        # image_key information
        with open_hdf5(url.file_path()) as h5s:
            dataset = h5s[url.data_path()]
            grp_parent = dataset.parent
            if grp_parent is not None and NXtomoScan.is_nxdetector(grp_parent):
                if "image_key" in grp_parent:
                    return h5py_read_dataset(grp_parent["image_key"])
        return None

    def _getSlices(self, image_key_values, serie_index):
        n_serie = -1
        start = None
        is_in_a_serie = False
        for i_frame, frame in enumerate(image_key_values):
            if frame == self._type.value and not is_in_a_serie:
                n_serie += 1
                is_in_a_serie = True
                if n_serie == serie_index:
                    start = i_frame
            elif frame != self._type.value and is_in_a_serie:
                is_in_a_serie = False
                if n_serie == serie_index:
                    return slice(start, i_frame)
        if start is not None:
            return slice(start, len(image_key_values))
        return None

    def getUrl(self):
        try:
            url_no_slices = DataUrl(path=self._qle.text())
            serie_index = self._optionsCB.currentText()
            if serie_index in ("", None):
                return None
            if serie_index == "full dataset":
                slices = None
            else:
                serie_index = int(serie_index.split("n°")[1])
                slices = self._getSlices(
                    image_key_values=self._getImageKey(url_no_slices),
                    serie_index=serie_index,
                )
            data_path = url_no_slices.data_path()
            url_slices = slices
            if url_slices is not None:
                url_slices = list(range(slices.start, slices.stop))
            url = DataUrl(
                file_path=url_no_slices.file_path(),
                data_path=data_path,
                data_slice=url_slices,
                scheme="silx",
            )
        except Exception as e:
            _logger.warning(f"Fail to create url. Reason is {e}")
            return None
        else:
            return url


class DarkRefPatchWidget(qt.QWidget):
    """
    Widget to add dark and flat
    """

    sigConfigurationChanged = qt.Signal()
    """Signal emit when the configuration changed (url update)"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())

        # start dark
        self._sdCB = qt.QCheckBox("start dark", self)
        self.layout().addWidget(self._sdCB, 0, 0, 1, 1)
        self._sdQLE = _DarkOrFlatUrl(self, ImageKey.DARK_FIELD, when="start")
        self.layout().addWidget(self._sdQLE, 0, 2, 1, 1)
        self._selectSD = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectSD, 0, 3, 1, 1)

        # start flat
        self._sfCB = qt.QCheckBox("start flat", self)
        self.layout().addWidget(self._sfCB, 1, 0, 1, 1)
        self._sfQLE = _DarkOrFlatUrl(self, ImageKey.FLAT_FIELD, when="start")
        self.layout().addWidget(self._sfQLE, 1, 2, 1, 1)
        self._selectSF = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectSF, 1, 3, 1, 1)

        # end dark
        self._edCB = qt.QCheckBox("end dark ", self)
        self.layout().addWidget(self._edCB, 2, 0, 1, 1)
        self._edQLE = _DarkOrFlatUrl(self, ImageKey.DARK_FIELD, when="end")
        self.layout().addWidget(self._edQLE, 2, 2, 1, 1)
        self._selectED = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectED, 2, 3, 1, 1)

        # end flat
        self._efCB = qt.QCheckBox("end flat", self)
        self.layout().addWidget(self._efCB, 3, 0, 1, 1)
        self._efQLE = _DarkOrFlatUrl(self, ImageKey.FLAT_FIELD, when="end")
        self.layout().addWidget(self._efQLE, 3, 2, 1, 1)
        self._selectEF = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectEF, 3, 3, 1, 1)

        # signal / slot connection
        # connect select button
        self._selectSD.released.connect(self._selectSDDataset)
        self._selectSF.released.connect(self._selectSFDataset)
        self._selectED.released.connect(self._selectEDDataset)
        self._selectEF.released.connect(self._selectEFDataset)

        # set up
        for widget_ in (
            self._sdQLE,
            self._selectSD,
            self._edQLE,
            self._selectED,
            self._sfQLE,
            self._selectSF,
            self._efQLE,
            self._selectEF,
        ):
            widget_.setEnabled(False)

        # connect checkbox buttons
        self._sdCB.toggled.connect(self._toggleSD)
        self._sfCB.toggled.connect(self._toggleSF)
        self._edCB.toggled.connect(self._toggleED)
        self._efCB.toggled.connect(self._toggleEF)

        # connect QLE modifications
        self._sdQLE.editingFinished.connect(self.sigConfigurationChanged)
        self._sdCB.toggled.connect(self.sigConfigurationChanged)
        self._sfQLE.editingFinished.connect(self.sigConfigurationChanged)
        self._sfCB.toggled.connect(self.sigConfigurationChanged)
        self._efQLE.editingFinished.connect(self.sigConfigurationChanged)
        self._efCB.toggled.connect(self.sigConfigurationChanged)
        self._edQLE.editingFinished.connect(self.sigConfigurationChanged)
        self._edCB.toggled.connect(self.sigConfigurationChanged)

    def _selectSDDataset(self):
        self._selectDataset("start dark", self._sdQLE)

    def _selectSFDataset(self):
        self._selectDataset("start flat", self._sfQLE)

    def _selectEDDataset(self):
        self._selectDataset("end dark", self._edQLE)

    def _selectEFDataset(self):
        self._selectDataset("end flat", self._efQLE)

    def _updateEnable(self, toggled, widgets):
        for w in widgets:
            w.setEnabled(toggled)

    def _toggleSD(self, toggled):
        self._updateEnable(toggled=toggled, widgets=(self._sdQLE, self._selectSD))

    def _toggleSF(self, toggled):
        self._updateEnable(toggled=toggled, widgets=(self._sfQLE, self._selectSF))

    def _toggleED(self, toggled):
        self._updateEnable(toggled=toggled, widgets=(self._edQLE, self._selectED))

    def _toggleEF(self, toggled):
        self._updateEnable(toggled=toggled, widgets=(self._efQLE, self._selectEF))

    def _selectDataset(self, what, outputQLE):
        # we can select a dataset or an entry. If this is an entry then we
        # will copy the target dark / flat...
        # ou ajouter un from: to
        assert isinstance(outputQLE, _DarkOrFlatUrl)
        dialog = DataFileDialog()
        dialog.setWindowTitle(what)
        dialog.setDirectory(get_default_directory())

        if dialog.exec_() and dialog.selectedDataUrl() is not None:
            old = outputQLE.blockSignals(True)
            url = dialog.selectedDataUrl()
            try:
                outputQLE.setUrl(url)
            except Exception as e:
                _logger.error(e)
            outputQLE.blockSignals(old)
            outputQLE.editingFinished.emit()

    def getStartDarkUrl(self) -> Union[None, DataUrl]:
        """

        :return: url defined by the user for patching the start darks
        :rtype:  Union[None, DataUrl]
        """
        if self._sdCB.isChecked():
            return self._sdQLE.getUrl()
        else:
            return None

    def getStartDarkIndex(self):
        return self._sdQLE.getSerieIndex()

    def setStartDarkUrl(self, url, serie_index=None) -> None:
        """

        :param DataUrl url:
        :param Union[int,None] serie_index: index of the serie to set
        """
        self._sdCB.setChecked(True)
        self._sdQLE.setUrl(url)
        if serie_index is not None:
            self._sdQLE.setSerie(serie_index)

    def getStartFlatUrl(self) -> Union[None, DataUrl]:
        """

        :return: url defined by the user for patching the start flats
        :rtype:  Union[None, DataUrl]
        """
        if self._sfCB.isChecked():
            return self._sfQLE.getUrl()
        else:
            return None

    def getStartFlatIndex(self):
        return self._sfQLE.getSerieIndex()

    def setStartFlatUrl(self, url, serie_index=None) -> None:
        """

        :param DataUrl url:
        :param Union[int,None] serie_index: index of the serie to set
        """
        self._sfCB.setChecked(True)
        self._sfQLE.setUrl(url)
        if serie_index is not None:
            self._sfQLE.setSerie(serie_index)

    def getEndDarkUrl(self) -> Union[None, DataUrl]:
        """

        :return: url defined by the user for patching the end darks
        :rtype:  Union[None, DataUrl]
        """
        if self._edCB.isChecked():
            return self._edQLE.getUrl()
        else:
            return None

    def getEndDarkIndex(self):
        return self._edQLE.getSerieIndex()

    def setEndDarkUrl(self, url, serie_index=None) -> None:
        """

        :param DataUrl url:
        :param Union[int,None] serie_index: index of the serie to set
        """
        self._edCB.setChecked(True)
        self._edQLE.setUrl(url)
        if serie_index is not None:
            self._edQLE.setSerie(serie_index)

    def getEndFlatUrl(self) -> Union[None, DataUrl]:
        """

        :return: url defined by the user for patching the end flats
        :rtype:  Union[None, DataUrl]
        """
        if self._efCB.isChecked():
            return self._efQLE.getUrl()
        else:
            return None

    def getEndFlatIndex(self):
        return self._efQLE.getSerieIndex()

    def setEndFlatUrl(self, url, serie_index=None) -> None:
        """

        :param DataUrl url:
        :param Union[int,None] serie_index: index of the serie to set
        """
        self._efCB.setChecked(True)
        self._efQLE.setUrl(url)
        if serie_index is not None:
            self._efQLE.setSerie(serie_index)

    def clear(self):
        for cb in self._sdCB, self._edCB, self._efCB, self._sfCB:
            cb.setChecked(False)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/04/2019"


import os
import time

import pytest

from tomwer.core.process.control.datawatcher import DataWatcher
from tomwer.core.process.control.datawatcher.status import DET_END_XML, PARSE_INFO_FILE
from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.utils.threads import LoopThread


@pytest.mark.skipif(os.name != "posix", reason="not tested under windows yet")
@pytest.mark.parametrize("det_method", (PARSE_INFO_FILE, DET_END_XML))
def test_data_watcher_io(tmp_path, det_method):
    """Test inputs and outputs types of the handler functions"""
    scan_folder = tmp_path / "folder" / "my_scan"
    data_watcher_process = DataWatcher()
    data_watcher_process.setWaitTimeBtwLoop(1)
    data_watcher_process.setObsMethod(det_method)

    MockEDF.mockScan(
        scanID=str(scan_folder), nRadio=10, nRecons=1, nPagRecons=4, dim=10
    )
    data_watcher_process.setFolderObserved(str(tmp_path / "folder"))
    data_watcher_process.clear_output_values()
    data_watcher_process.set_serialize_output_data(True)
    LoopThread.quitEvent.clear()
    data_watcher_process.start()

    # note: this quitEvent is nasty. But we don't want to spend time on the (EDF) data watcher which is not the future

    timeout = 8
    while data_watcher_process.get_output_value("data") is None and timeout > 0:
        _delta = 0.5
        time.sleep(_delta)
        timeout -= _delta

    def quit_test():
        data_watcher_process.stop()
        data_watcher_process.waitForObservationFinished()
        import gc

        gc.collect()
        LoopThread.quitEvent.clear()

    if timeout <= 0:
        quit_test()
        raise TimeoutError("timeout expire")

    out = data_watcher_process.get_output_value("data")
    assert isinstance(out, dict)
    quit_test()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/07/2022"


import os
import pathlib
from datetime import datetime

from processview.core.dataset import Dataset, DatasetIdentifier
from tomoscan.esrf.identifier.tiffidentifier import (
    MultiTiffVolumeIdentifier as _MultiTiffVolumeIdentifier,
)
from tomoscan.esrf.identifier.tiffidentifier import (
    TIFFVolumeIdentifier as _TIFFVolumeIdentifier,
)
from tomoscan.esrf.volume.tiffvolume import MultiTIFFVolume as _MultiTIFFVolume
from tomoscan.esrf.volume.tiffvolume import TIFFVolume as _TIFFVolume

from tomwer.core.volume.volumebase import TomwerVolumeBase


class TIFFVolumeIdentifier(_TIFFVolumeIdentifier, DatasetIdentifier):
    def __init__(self, object, folder, file_prefix, metadata=None):
        super().__init__(object, folder, file_prefix)
        DatasetIdentifier.__init__(self, TIFFVolume.from_identifier, metadata=metadata)

    @staticmethod
    def from_str(identifier):
        return _TIFFVolumeIdentifier._from_str_to_single_frame_identifier(
            identifier=identifier,
            SingleFrameIdentifierClass=TIFFVolumeIdentifier,
            ObjClass=TIFFVolume,
        )

    def long_description(self) -> str:
        """used for processview header tooltip for now"""
        return self.to_str()


class MultiTiffVolumeIdentifier(_MultiTiffVolumeIdentifier, DatasetIdentifier):
    def __init__(self, object, tiff_file, metadata=None):
        super().__init__(object, tiff_file)
        DatasetIdentifier.__init__(self, TIFFVolume.from_identifier, metadata=metadata)

    @staticmethod
    def from_str(identifier):
        identifier_no_scheme = identifier.split(":")[-1]
        tiff_file = identifier_no_scheme
        return MultiTiffVolumeIdentifier(object=TIFFVolume, tiff_file=tiff_file)


class TIFFVolume(_TIFFVolume, TomwerVolumeBase, Dataset):
    @staticmethod
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, TIFFVolumeIdentifier):
            raise TypeError(
                f"identifier should be an instance of {TIFFVolumeIdentifier} not {type(identifier)}"
            )
        return TIFFVolume(
            folder=identifier.folder,
            volume_basename=identifier.file_prefix,
        )

    def get_identifier(self) -> TIFFVolumeIdentifier:
        if self.url is None:
            raise ValueError("no file_path provided. Cannot provide an identifier")

        try:
            stat = pathlib.Path(self.url.file_path()).stat()
        except Exception:
            stat = None

        return TIFFVolumeIdentifier(
            object=self,
            folder=self.url.file_path(),
            file_prefix=self._volume_basename,
            metadata={
                "name": self.url.file_path(),
                "creation_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
                "modification_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
            },
        )

    def volume_data_parent_folder(self):
        if self.data_url is None:
            raise ValueError("data_url doesn't exists")
        else:
            return os.path.dirname(self.data_url.file_path())


class MultiTIFFVolume(_MultiTIFFVolume, TomwerVolumeBase, Dataset):
    @staticmethod
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, MultiTiffVolumeIdentifier):
            raise TypeError(
                f"identifier should be an instance of {MultiTiffVolumeIdentifier}"
            )
        return MultiTIFFVolume(
            file_path=identifier.file_path,
        )

    def get_identifier(self) -> MultiTiffVolumeIdentifier:
        if self.url is None:
            raise ValueError("no file_path provided. Cannot provide an identifier")

        try:
            stat = pathlib.Path(self.url.file_path()).stat()
        except Exception:
            stat = None

        return MultiTiffVolumeIdentifier(
            object=self,
            tiff_file=self.url.file_path(),
            metadata={
                "name": self.url.file_path(),
                "creation_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
                "modification_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
            },
        )

    def volume_data_parent_folder(self):
        if self.data_url is None:
            raise ValueError("data_url doesn't exists")
        else:
            return os.path.dirname(self.data_url.file_path())



import logging

from silx.gui import qt

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output, OWBaseWidget
from tomoscan.serie import Serie

from tomwer.core.tomwer_object import TomwerObject
from tomwer.gui.stitching.stitching import ZStitchingWindow

_logger = logging.getLogger(__name__)


class ZStitchingConfigOW(OWBaseWidget):
    """
    Widget to create a stitching configuration to be used with nabu-stitching
    """

    name = "z-stitching config"
    id = "orange.widgets.tomwer.stitching.ZStitchingConfigOW.ZStitchingConfigOW"
    description = (
        "Interface to create / edit a z-stitching to be used with nabu-stitching"
    )
    icon = "icons/zstitching_icon.svg"
    priority = 60
    keywords = [
        "hdf5",
        "tomwer",
        "NXTomo",
        "stitching",
        "z-stitching",
        "z-serie",
        "zserie",
        "pre-processing",
        "radios",
        "projections",
    ]

    want_main_area = True
    resizing_enabled = True
    want_control_area = False

    _ewoks_default_inputs = Setting({})

    class Inputs:
        tomo_obj = Input(
            name="tomo_obj",
            type=TomwerObject,
            doc="one scan to be process",
            default=True,
            multiple=True,
        )

        serie = Input(
            "serie",
            type=Serie,
            doc="serie of scan or volumes to be stitched together",
            default=False,
            multiple=False,
        )

    class Outputs:
        stitching_config = Output(
            name="stitching configuration",
            type=dict,
            doc="configuration to stitch together tomo object as a dictionary",
        )

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(parent=parent, *args, **kwargs)

        _layout = gui.vBox(self.mainArea, self.name).layout()
        self.widget = ZStitchingWindow(parent=self)
        _layout.addWidget(self.widget)

        # connect signal / slot
        self.widget.sigChanged.connect(self._storeSettings)

        # load settings
        self.setConfiguration(self._ewoks_default_inputs)

    def setConfiguration(self, config: dict) -> None:
        self.widget.setConfiguration(config=config)

    def getConfiguration(self) -> dict:
        return self.widget.getConfiguration()

    def _storeSettings(self):
        self._ewoks_default_inputs = self.widget.getConfiguration()

    @Inputs.tomo_obj
    def addTomoObj(self, tomo_obj, *args, **kwargs):
        self.widget.addTomoObj(tomo_obj)

    @Inputs.serie
    def setSerie(self, serie, *args, **kwargs):
        self.widget.setSerie(serie)
        self._storeSettings()

    def validate(self):
        config = self.widget.getConfiguration()
        assert isinstance(config, dict)
        self.Outputs.stitching_config.send(config)

    def keyPressEvent(self, event):
        """
        To shortcut orange and make sure the `F5` <=> refresh stitching preview
        """
        if event.key() == qt.Qt.Key_F5:
            self.widget._trigger_update_preview()
        else:
            super().keyPressEvent(event)

    def close(self):
        self.widget.close()
        super().close()
